@@ -11,6 +11,9 @@
 #ifndef __NR_getpgid
 #define __NR_getpgid 132
 #endif
+#ifndef __NR_capget
+#define __NR_capget 184
+#endif
 #ifndef __NR_gettid
 #define __NR_gettid 224
 #endif
@@ -26,3 +29,6 @@
 #ifndef __NR_setns
 #define __NR_setns 346
 #endif
+#ifndef __NR_seccomp
+#define __NR_seccomp 354
+#endif
